@@ -33,6 +33,8 @@
 
 #include "config.h"
 
+#define _GNU_SOURCE /* need sighandler_t */
+
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/stat.h>
@@ -47,6 +49,8 @@
 #include <termios.h>
 #include <string.h>
 #include <curses.h>
+#include <signal.h>
+#include <errno.h>
 
 #include "dgamelaunch.h"
 #include "ttyplay.h"
@@ -54,7 +58,22 @@
 #include "io.h"
 #include "stripgfx.h"
 
+#ifdef __MACH__
+typedef void (*sighandler_t)(int);
+#endif
+
 int stripped = NO_GRAPHICS;
+static int got_sigwinch = 0;
+
+static int term_resizex = -1;
+static int term_resizey = -1;
+
+void
+ttyplay_sigwinch_func(int sig)
+{
+    signal(SIGWINCH, ttyplay_sigwinch_func);
+    got_sigwinch = 1;
+}
 
 struct timeval
 timeval_diff (struct timeval tv1, struct timeval tv2)
@@ -103,11 +122,84 @@ ttynowait (struct timeval prev, struct timeval cur, double speed)
   return 0;                     /* Speed isn't important. */
 }
 
+int
+kbhit(void)
+{
+    int i = 0;
+    nodelay(stdscr, TRUE);
+    timeout(0);
+    i = wgetch(stdscr);
+    nodelay(stdscr, FALSE);
+
+    if (i == -1)
+        i = 0;
+    else
+        ungetch(i);
+    return (i);
+}
+
+int
+ttyplay_keyboard_action(int c)
+{
+    struct termios t;
+    switch (c)
+    {
+    case 'q':
+        return READ_QUIT;
+    case 'r':
+        if (term_resizex > 0 && term_resizey > 0) {
+            printf ("\033[8;%d;%dt", term_resizey, term_resizex);
+            return READ_RESTART;
+        }
+        break;
+    case 's':
+        switch (stripped)
+        {
+	case NO_GRAPHICS: populate_gfx_array ((stripped = DEC_GRAPHICS)); break;
+	case DEC_GRAPHICS: populate_gfx_array ((stripped = IBM_GRAPHICS)); break;
+	case IBM_GRAPHICS: populate_gfx_array ((stripped = NO_GRAPHICS)); break;
+        }
+        return READ_RESTART;
+
+    case 'm':
+        tcgetattr (0, &t);
+        if (!loggedin)
+        {
+            initcurses();
+            loginprompt(1);
+        }
+        if (loggedin)
+        {
+            initcurses ();
+            domailuser (chosen_name);
+        }
+        endwin ();
+        tcsetattr (0, TCSANOW, &t);
+        return READ_RESTART;
+    case '?':
+        tcgetattr (0, &t);
+        initcurses();
+        (void) runmenuloop(dgl_find_menu("watchmenu_help"));
+        endwin ();
+        tcsetattr (0, TCSANOW, &t);
+        return READ_RESTART;
+    }
+    return (READ_DATA);
+}
+
 int
 ttyread (FILE * fp, Header * h, char **buf, int pread)
 {
   long offset;
 
+  if (kbhit())
+      {
+	  const int c = dgl_getch();
+	  const int action = ttyplay_keyboard_action(c);
+	  if (action != READ_DATA)
+	      return (action);
+      }
+
   /* do this BEFORE header read, hlen bug */
   offset = ftell (fp);
 
@@ -124,7 +216,7 @@ ttyread (FILE * fp, Header * h, char **buf, int pread)
       exit (-21);
     }
 
-  *buf = malloc (h->len);
+  *buf = malloc (h->len + 1);
   if (*buf == NULL)
     {
       perror ("malloc");
@@ -136,6 +228,7 @@ ttyread (FILE * fp, Header * h, char **buf, int pread)
       fseek (fp, offset, SEEK_SET);
       return READ_EOF;
     }
+  (*buf)[h->len] = 0;
   return READ_DATA;
 }
 
@@ -151,8 +244,8 @@ ttypread (FILE * fp, Header * h, char **buf, int pread)
   struct timeval origw = { 0, 100000 };
   int counter = 0;
   fd_set readfs;
-  struct termios t;
   int doread = 0;
+  int action = READ_DATA;
 
 #ifdef HAVE_KQUEUE
   if (kq == -1)
@@ -167,8 +260,9 @@ ttypread (FILE * fp, Header * h, char **buf, int pread)
   /*
    * Read persistently just like tail -f.
    */
-  while (ttyread (fp, h, buf, 1) == READ_EOF)
+  while ((action = ttyread (fp, h, buf, 1)) == READ_EOF)
     {
+      idle_alarm_reset();
       fflush(stdout);
       clearerr (fp);
 #ifdef HAVE_KQUEUE
@@ -215,49 +309,24 @@ ttypread (FILE * fp, Header * h, char **buf, int pread)
       }
       if (n == -1)
 	{
-	  printf("select()/kevent() failed.\n");
-	  exit (1);
+	    if ((errno == EINTR) && got_sigwinch) {
+		got_sigwinch = 0;
+		return READ_RESTART;
+	    } else {
+		printf("select()/kevent() failed.\n");
+		exit (1);
+	    }
 	}
       if (doread)
         {                       /* user hits a character? */
-          char c;
-          read (STDIN_FILENO, &c, 1); /* drain the character */
+	  const int c = dgl_getch();
+	  action = ttyplay_keyboard_action(c);
+	  if (action != READ_DATA)
+	      return action;
 
-          switch (c)
-            {
-            case 'q':
-              return READ_EOF;
-              break;
-	    case 's':
-	      switch (stripped)
-	      {
-		case NO_GRAPHICS: populate_gfx_array ((stripped = DEC_GRAPHICS)); break;
-		case DEC_GRAPHICS: populate_gfx_array ((stripped = IBM_GRAPHICS)); break;
-		case IBM_GRAPHICS: populate_gfx_array ((stripped = NO_GRAPHICS)); break;
-	      }
-	      return READ_RESTART;
-	      break;
-
-            case 'm':
-	      tcgetattr (0, &t);
-	      if (!loggedin)
-	      {
-		initcurses();
-		loginprompt(1);
-	      }
-              if (loggedin)
-	      {
-		initcurses ();
-		domailuser (chosen_name);
-	      }
-              endwin ();
-	      tcsetattr (0, TCSANOW, &t);
-              return READ_RESTART;
-              break;
-            }
         }
     }
-  return READ_DATA;
+  return (action);
 }
 
 void
@@ -318,54 +387,69 @@ ttyplay (FILE * fp, double speed, ReadFunc read_func,
   return r;
 }
 
+static off_t
+find_last_string_in_file(FILE * fp, const char *seq)
+{
+    char buf[512];
+    struct stat mystat;
+    off_t offset = 0L;
+    const long readsz = sizeof(buf);
+    int bytes_read = 0;
+    const int seqlen = strlen(seq);
+    const char *reset_pos = seq + seqlen - 1;
+    const char *match_pos = reset_pos;
+
+    fstat(fileno (fp), &mystat);
+    offset = mystat.st_size - readsz;
+    if (offset < 0)
+        offset = 0;
+    while (1)
+    {
+        const char *search_pos = 0;
+
+        fseeko(fp, offset, SEEK_SET);
+        bytes_read = fread(buf, 1, readsz, fp);
+
+        if (bytes_read <= 0)
+            break;
+
+        search_pos = buf + bytes_read - 1;
+        while (search_pos >= buf)
+        {
+            int matched = *search_pos == *match_pos;
+            if (!matched && match_pos != reset_pos)
+            {
+                match_pos = reset_pos;
+                matched = *search_pos == *match_pos;
+            }
+            if (matched)
+            {
+                if (match_pos == seq)
+                    return offset + (search_pos - buf);
+                --match_pos;
+            }
+            --search_pos;
+        }
+
+        // If we've reached the start of the file, exit.
+        if (!offset)
+            break;
+
+        offset -= readsz;
+        if (offset < 0)
+            offset = 0;
+    }
+
+    return 0;
+}
+
 static off_t
 find_seek_offset_clrscr (FILE * fp)
 {
   off_t raw_seek_offset = 0;
   off_t seek_offset_clrscr;
-  char *buf;
-  struct stat mystat;
-  int state = 0;
-  int i;
-  int bytesread;
 
-  fstat (fileno (fp), &mystat);
-  buf = malloc (mystat.st_size);
-  fseek (fp, 0, SEEK_SET);
-  bytesread = fread (buf, 1, mystat.st_size, fp);
-
-  /* one byte at at time sucks, but is a simple hack for the temp
-   * being to avoid looking for wraparounds */
-  for (i = 0; i < bytesread; i++)
-    {
-      if (buf[i] == 0x1b)
-        {
-          state = 1;
-        }
-      else if ((buf[i] == 0x5b) && (state == 1))
-        {
-          state = 2;
-        }
-      else if ((buf[i] == 0x32) && (state == 2))
-        {
-          state = 3;
-        }
-      else if ((buf[i] == 0x4a) && ((state == 2) || (state == 3)))
-        {
-          state = 4;
-        }
-      else
-        {
-          state = 0;
-        }
-
-      if (state == 4)
-        {
-          raw_seek_offset = i - 2;
-        }
-    }
-
-  free (buf);
+  raw_seek_offset = find_last_string_in_file(fp, "\033[2J");
 
   seek_offset_clrscr = 0;
   /* now find last filepos that is less than seek offset */
@@ -374,18 +458,20 @@ find_seek_offset_clrscr (FILE * fp)
     {
       char *buf;
       Header h;
+      long offset;
 
       if (ttyread (fp, &h, &buf, 0) != READ_DATA)
         {
           break;
         }
 
-      if (ftell (fp) < raw_seek_offset)
-        {
-          seek_offset_clrscr = ftell (fp);
-        }
-
       free (buf);
+
+      offset = ftell(fp);
+      if (offset < raw_seek_offset)
+          seek_offset_clrscr = ftell (fp);
+      else
+          break;
     }
 
   return seek_offset_clrscr;
@@ -412,27 +498,29 @@ void
 ttypeek (FILE * fp, double speed)
 {
   int r;
-
   do
   {
     setvbuf (fp, NULL, _IOFBF, 0);
-    ttyplay (fp, 0, ttyread, ttywrite, ttynowait, find_seek_offset_clrscr (fp));
-    clearerr (fp);
-    setvbuf (fp, NULL, _IONBF, 0);
-    fflush (stdout);
-    r = ttyplay (fp, speed, ttypread, ttywrite, ttynowait, -1);
+    r = ttyplay(fp, 0, ttyread, ttywrite, ttynowait, find_seek_offset_clrscr(fp));
+    if (r == READ_EOF) {
+	clearerr (fp);
+        setvbuf (fp, NULL, _IONBF, 0);
+        fflush (stdout);
+        r = ttyplay (fp, speed, ttypread, ttywrite, ttynowait, -1);
+    }
   } while (r == READ_RESTART);
 }
 
 
 int
-ttyplay_main (char *ttyfile, int mode)
+ttyplay_main (char *ttyfile, int mode, int resizex, int resizey)
 {
   double speed = 1.0;
   ReadFunc read_func = ttyread;
   WaitFunc wait_func = ttywait;
   FILE *input = stdin;
   struct termios old, new;
+  sighandler_t old_sigwinch;
 
   populate_gfx_array (stripped);
 
@@ -445,6 +533,15 @@ ttyplay_main (char *ttyfile, int mode)
   new.c_cc[VTIME] = 0;
   tcsetattr (0, TCSANOW, &new); /* Make it current */
 
+
+  if (resizex > 0 && resizey > 0) {
+      term_resizex = resizex;
+      term_resizey = resizey;
+  }
+
+  got_sigwinch = 0;
+  old_sigwinch = signal(SIGWINCH, ttyplay_sigwinch_func);
+
   if (mode == 1)
     ttypeek (input, speed);
   else
@@ -453,5 +550,10 @@ ttyplay_main (char *ttyfile, int mode)
   tcsetattr (0, TCSANOW, &old); /* Return terminal state */
   fclose (input);
 
+  if (old_sigwinch != SIG_ERR)
+      signal(SIGWINCH, old_sigwinch);
+
+  term_resizex = term_resizey = -1;
+
   return 0;
 }
